"""
Functions for actually downloading a photo -- you get the URL from
the Flickr API, then download the image itself here.
"""

from pathlib import Path
import uuid

import httpx
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception


client = httpx.Client(headers={"User-Agent": "flickr-photos-api"})


def is_retryable(exc: BaseException) -> bool:
    """
    Returns True if this is an exception we can safely retry (i.e. flaky
    or transient errors that might return a different result),or
    False otherwise.
    """
    if isinstance(exc, httpx.HTTPStatusError) and exc.response.status_code in {
        403,
        429,
    }:
        return True  # pragma: no cover

    if isinstance(exc, httpx.RemoteProtocolError):
        return True  # pragma: no cover

    # We can retry a download if it timed out.
    #
    # There's no easy way to simulate this in tests, so we just don't
    # test this branch -- but it's simple enough not to be a big issue.
    if isinstance(exc, (httpx.ConnectTimeout, httpx.ReadTimeout)):  # pragma: no cover
        return True

    return False


@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=60),
    retry=retry_if_exception(is_retryable),
)
def download_photo(url: str, out_path: Path) -> None:
    """
    Download a photo from Flickr.com.

    This function will retry the download a certain number of times
    if it fails, e.g. if Flickr returns an error or if the download
    is rate-limited.
    """
    out_path.parent.mkdir(exist_ok=True, parents=True)

    tmp_path = out_path.with_suffix(f".{uuid.uuid4()}.tmp")

    # We use a streaming response from HTTPX because some Flickr files
    # can be very big, e.g. original video files.  We don't need or
    # want to buffer the whole thing into memory.
    with client.stream("GET", url) as resp:
        resp.raise_for_status()

        with open(tmp_path, "xb") as out_file:
            for data in resp.iter_bytes():
                out_file.write(data)

    tmp_path.rename(out_path)



"""
Tests for ``flickr_photos_api.downloader``.
"""

from collections.abc import Iterator
import hashlib
from pathlib import Path
import time

import httpx
import pytest
import vcr

from flickr_photos_api import download_photo


@pytest.fixture
def vcr_cassette(cassette_name: str) -> Iterator[str]:
    """
    A basic pytest fixture to save responses to the cassette directory,
    so we don't download the file every time.
    """
    with vcr.use_cassette(
        cassette_name,
        cassette_library_dir="tests/fixtures/cassettes",
        decode_compressed_response=True,
    ):
        yield cassette_name


def test_download_photo(vcr_cassette: str, tmp_path: Path) -> None:
    """
    Download a photo from Flickr and check it's downloaded correctly.
    """
    out_path = tmp_path / "53574198477.jpg"

    download_photo(
        url="https://live.staticflickr.com/65535/53574198477_fba34d20ca_c_d.jpg",
        out_path=out_path,
    )

    assert out_path.exists()
    assert out_path.stat().st_size == 126058
    assert (
        hashlib.md5(out_path.read_bytes()).hexdigest()
        == "392b2e74d29ff90bb707658d422d14ad"
    )


def test_not_found_is_error(vcr_cassette: str, tmp_path: Path) -> None:
    """
    Trying to fetch a Flickr URL that doesn't exist throws an immediate
    404 error.
    """
    t0 = time.time()

    with pytest.raises(httpx.HTTPStatusError):
        download_photo(
            url="https://live.staticflickr.com/65535/doesnotexist.jpg",
            out_path=tmp_path / "doesnotexist.jpg",
        )

    # Check that less than 5 seconds elapsed -- we weren't waiting for
    # the library to retry anything.
    assert time.time() - t0 < 5
